////////////////////////////////////////////////////////////////
// Line Trace Robbot Simulator
//        
//        Copyright(C) Nobuyuki Umetani 2008
////////////////////////////////////////////////////////////////

#include <GL/glut.h>
#include <stdio.h>
#include <math.h>

#define PI  3.14
#define WIDTH 1.9
#define H_WIDTH 0.95

double rotation1;
double rotation2;

double circle[10][3];
double angle[10][2];
double vertex[22][3];
double radius[10];

double robo_x_position;
double robo_y_position;
double robo_r_tire;
double robo_l_tire;
double robo_velocity[2];
double robo_theta;
double robo_omega;
double robo_tire_scale;
int robo_sencer[3];

GLdouble robo_vertex[][3] = {
    { -5.0, -5.0, 1.0 }, { 5.0, -5.0, 1.0 },
    { 5.0, 5.0, 1.0 }, { -5.0, 5.0, 1.0 },
    { -5.0, -5.0, 8.0 }, { 5.0, -5.0, 8.0 },
    { 5.0, 5.0, 8.0 }, { -5.0, 5.0, 8.0 },
	{ 6.0, 6.0, 1.0 }, { 9.0, 6.0, 1.0},
	{ 9.0, -6.0, 1.0}, { 6.0, -6.0, 1.0},
	{ 8.0, 4.0, 2.0}, {8.0, 0.0, 2.0},
	{ 8.0, -4.0, 2.0}
};

int robo_edge[][2] = {
    { 0, 1 }, { 1, 2 }, { 2, 3 }, { 3, 0 },
    { 4, 5 }, { 5, 6 }, { 6, 7 }, { 7, 4 },
    { 0, 4 }, { 1, 5 }, { 2, 6 }, { 3, 7 }, 
	{ 8, 9 }, { 9, 10}, {10, 11}, {11, 8}
};

GLdouble robo_normal[][3] = {
  { 0.0, 0.0,-1.0 },
  { 1.0, 0.0, 0.0 },
  { 0.0, 0.0, 1.0 },
  {-1.0, 0.0, 0.0 },
  { 0.0,-1.0, 0.0 },
  { 0.0, 1.0, 0.0 },
  { 0.0, 0.0, 1.0 }
};

int robo_face[][4] = {
  { 0, 1, 2, 3 },
  { 1, 5, 6, 2 },
  { 5, 4, 7, 6 },
  { 4, 0, 3, 7 },
  { 4, 5, 1, 0 },
  { 3, 2, 6, 7 },
  { 8, 9, 10, 11}
};




void initialize_scene(void)
{
	int i;
	circle[0][0] = 60.0;
	circle[0][1] = 60.0;
	circle[0][2] = 0.0;

	circle[1][0] = 60 - 40*cos(PI/8.0) + 10*sin(PI/8.0) - 15*cos(PI/8.0);
	circle[1][1] = 60 + 40*sin(PI/8.0) + 10*cos(PI/8.0) + 15*sin(PI/8.0);
	circle[1][2] = 0.0;

	circle[2][0] = 60 - 40*cos(PI/8.0) + 30*cos(PI/8.0);
	circle[2][1] = 60 + 40*sin(PI/8.0) + 10*cos(PI/8.0)*2 + 15*sin(PI/8.0)*2 + 30*sin(PI/8.0);
	circle[2][2] = 0.0;

	circle[3][0] = 60.0;
	circle[3][1] = 60.0;
	circle[3][2] = 0.0;

	circle[4][0] = 60 + 20*cos(PI/8.0) + 86*sin(PI/8.0) + 20*cos(PI/8.0);
	circle[4][1] = 60 - 20*sin(PI/8.0) + 86*cos(PI/8.0) - 20*sin(PI/8.0);
	circle[4][2] = 0.0;

	circle[5][0] = 60 + 20*cos(PI/8.0) + 86*sin(PI/8.0) + 20*cos(PI/8.0) + 10;
	circle[5][1] = 60 - 20*sin(PI/8.0) + 86*cos(PI/8.0) - 20*sin(PI/8.0) + 6;
	circle[5][2] = 0.0;

	circle[6][0] = 60 + 20*cos(PI/8.0) + 86*sin(PI/8.0) + 20*cos(PI/8.0) + 4;
	circle[6][1] = 60 - 20*sin(PI/8.0) + 86*cos(PI/8.0) - 20*sin(PI/8.0) - 22;
	circle[6][2] = 0.0;

	circle[7][0] = 60 + 20*cos(PI/8.0) + 86*sin(PI/8.0) + 20*cos(PI/8.0) + 10;
	circle[7][1] = 60 - 20*sin(PI/8.0) + 86*cos(PI/8.0) - 20*sin(PI/8.0) - 50;
	circle[7][2] = 0.0;

	circle[8][0] = 60 + 20*cos(PI/8.0) + 86*sin(PI/8.0) + 20*cos(PI/8.0) + 10 - 60 + ( 60 - 20*sin(PI/8.0) + 86*cos(PI/8.0) - 20*sin(PI/8.0) - 50 - 35 )*cos(PI/6.0);
	circle[8][1] = ( 60 - 20*sin(PI/8.0) + 86*cos(PI/8.0) - 20*sin(PI/8.0) - 50 + 35 ) / 2;
	circle[8][2] = 0.0;

	circle[9][0] = 60 + 20*cos(PI/8.0) + 86*sin(PI/8.0) + 20*cos(PI/8.0) + 10;
	circle[9][1] = 34;
	circle[9][2] = 0.0;

	angle[0][0] = PI*3.0/2.0;	angle[0][1] = PI*7.0/8.0;
	angle[1][0] = -PI/8.0;	angle[1][1] = PI/8.0;
	angle[2][0] = PI*9.0/8.0;	angle[2][1] = -PI/4.0;
	angle[3][0] = PI*3.0/4.0;	angle[3][1] = PI*15.0/8.0;
	angle[4][0] = PI*7.0/8.0;	angle[4][1] = PI/2.0;
	angle[5][0] = PI/2.0;	angle[5][1] = -PI/2.0;
	angle[6][0] = PI/2.0;	angle[6][1] = PI*3.0/2.0;
	angle[7][0] = PI/2.0;	angle[7][1] = -PI*2.0/3.0;
	angle[8][0] = PI/3.0;	angle[8][1] = PI*5.0/3.0;
	angle[9][0] = PI*2.0/3.0;	angle[9][1] = -PI/2.0;

	radius[0] = 40.0;
	radius[1] = 15.0;
	radius[2] = 30.0;
	radius[3] = 20.0;
	radius[4] = 20.0;
	radius[5] = 14.0;
	radius[6] = 14.0;
	radius[7] = 14.0;
	radius[8] = 15.0;
	radius[9] = 14.0;

	vertex[0][0] = 95.0;
	vertex[0][1] = 20.0;
	vertex[0][2] = 0.1;

	for(i=0;i<10;i++){
		vertex[2*i+1][0] = circle[i][0] + radius[i]*cos(angle[i][0]);
		vertex[2*i+1][1] = circle[i][1] + radius[i]*sin(angle[i][0]);
		vertex[2*i+1][2] = 0.1;

		vertex[2*i+2][0] = circle[i][0] + radius[i]*cos(angle[i][1]);
		vertex[2*i+2][1] = circle[i][1] + radius[i]*sin(angle[i][1]);
		vertex[2*i+2][2] = 0.1;
	}

	vertex[21][0] = 95.0;
	vertex[21][1] = 20.0;
	vertex[21][2] = 0.1;

}

void initialize_robo(void){

	robo_x_position = 95.0;
	robo_y_position = 20.0;
	robo_theta = PI;
	robo_omega = 0.0;
	robo_r_tire = 00;
	robo_l_tire = 00;
	robo_sencer[0] = 0;
	robo_sencer[1] = 0;
	robo_sencer[2] = 0;
	robo_tire_scale = 0.008;
}

////////////////////////////////////////////////////////////////////
//NbN_ɂsNZ̃fvXlpāCEWoD
void  WindCo(
             double x, double y, double depth, 
			 double* wx, double* wy,double* wz){
        // 4*4spӂ
        GLdouble mvMatrix[16],pjMatrix[16];
        // r[|[gpz
        GLint viewport[4];
        // p[^擾D
        glGetIntegerv(GL_VIEWPORT, viewport);
        glGetDoublev(GL_MODELVIEW_MATRIX, mvMatrix);
        glGetDoublev(GL_PROJECTION_MATRIX, pjMatrix);
        //EW擾
        gluProject(x, y, depth, 
                mvMatrix,
                pjMatrix,
                viewport,
                wx,
                wy,
                wz);
}

void draw_course(void){
	int i;
	double d;

	glColor3d(1.0, 1.0, 1.0);
	glBegin(GL_QUAD_STRIP);
	for(i=0;i<10;i++){
		if(angle[i][0] < angle[i][1]){
			for(d = angle[i][0]; d < angle[i][1]; d += PI/180.0){
				glVertex3d(circle[i][0] + (radius[i] - H_WIDTH)*cos(d), circle[i][1] + (radius[i] - H_WIDTH)*sin(d), 0.0);
				glVertex3d(circle[i][0] + (radius[i] + H_WIDTH)*cos(d), circle[i][1] + (radius[i] + H_WIDTH)*sin(d), 0.0);
			}
			glVertex3d(circle[i][0] + (radius[i] - H_WIDTH)*cos(angle[i][1]), circle[i][1] + (radius[i] - H_WIDTH)*sin(angle[i][1]), 0.0);
			glVertex3d(circle[i][0] + (radius[i] + H_WIDTH)*cos(angle[i][1]), circle[i][1] + (radius[i] + H_WIDTH)*sin(angle[i][1]), 0.0);
		}
		else{
			for(d = angle[i][0]; d > angle[i][1]; d -= PI/180.0){
				glVertex3d(circle[i][0] + (radius[i] + H_WIDTH)*cos(d), circle[i][1] + (radius[i] + H_WIDTH)*sin(d), 0.0);
				glVertex3d(circle[i][0] + (radius[i] - H_WIDTH)*cos(d), circle[i][1] + (radius[i] - H_WIDTH)*sin(d), 0.0);
			}
			glVertex3d(circle[i][0] + (radius[i] + H_WIDTH)*cos(angle[i][1]), circle[i][1] + (radius[i] + H_WIDTH)*sin(angle[i][1]), 0.0);
			glVertex3d(circle[i][0] + (radius[i] - H_WIDTH)*cos(angle[i][1]), circle[i][1] + (radius[i] - H_WIDTH)*sin(angle[i][1]), 0.0);
		}
	}
	glVertex3d(circle[0][0] + (radius[0] + H_WIDTH)*cos(angle[0][0]), circle[0][1] + (radius[0] + H_WIDTH)*sin(angle[0][0]), 0.0);
	glVertex3d(circle[0][0] + (radius[0] - H_WIDTH)*cos(angle[0][0]), circle[0][1] + (radius[0] - H_WIDTH)*sin(angle[0][0]), 0.0);
	glEnd();
}

void draw_course_line(void){
	int i;
	double d;

	glColor3d(1.0, 1.0, 1.0);
	glBegin(GL_LINES);
	for(i=0;i<22;i++){
		glVertex3dv(vertex[i]);	//0
	}
	glEnd();

	for(i=0;i<10;i++){
		glBegin(GL_LINE_STRIP);
		if(angle[i][0] < angle[i][1]){
			for(d = angle[i][0]; d < angle[i][1]; d += PI/180.0){
				glVertex3d(circle[i][0] + radius[i]*cos(d), circle[i][1] + radius[i]*sin(d), 0.0);
			}
			glVertex3d(circle[i][0] + radius[i]*cos(angle[i][1]), circle[i][1] + radius[i]*sin(angle[i][1]), 0.0);
		}
		else{
			for(d = angle[i][0]; d > angle[i][1]; d -= PI/180.0){
				glVertex3d(circle[i][0] + radius[i]*cos(d), circle[i][1] + radius[i]*sin(d), 0.0);
			}
			glVertex3d(circle[i][0] + radius[i]*cos(angle[i][1]), circle[i][1] + radius[i]*sin(angle[i][1]), 0.0);
		}
		glEnd();
	}
}

int pick(double, double);

void draw_robo(void){
	int i;

	glPushMatrix();
	glTranslated(robo_x_position, robo_y_position, 0.0);
	glRotated( robo_theta*180.0/PI, 0.0, 0.0, 1.0);
	glColor3d(0.0, 0.0, 1.0);
	glBegin(GL_LINES);
	for (i = 0; i < 16; i++) {
		glVertex3dv(robo_vertex[robo_edge[i][0]]);
		glVertex3dv(robo_vertex[robo_edge[i][1]]);
	}
	glEnd();

	glBegin(GL_QUADS);
	for( i = 0;i < 7; i++ ){
		glVertex3dv(robo_vertex[robo_face[i][0]]);
		glVertex3dv(robo_vertex[robo_face[i][1]]);
		glVertex3dv(robo_vertex[robo_face[i][2]]);
		glVertex3dv(robo_vertex[robo_face[i][3]]);
	}
	glEnd();

	robo_sencer[0] = 0;
	robo_sencer[1] = 0;
	robo_sencer[2] = 0;

	if( pick(robo_vertex[12][0], robo_vertex[12][1] ) == 1 ){
		robo_sencer[0] = 1;
	}
	if( pick(robo_vertex[13][0], robo_vertex[13][1] ) == 1 ){
		robo_sencer[1] = 1;
	}
	if( pick(robo_vertex[14][0], robo_vertex[14][1] ) == 1 ){
		robo_sencer[2] = 1;
	}

	glPointSize(4);
	glBegin(GL_POINTS);
	if( robo_sencer[0] == 1 ){
		glColor3d(1.0, 0.0, 0.0);
	}
	else	glColor3d(0.2, 0.0, 0.0);
	glVertex3dv(robo_vertex[12]);

	if( robo_sencer[1] == 1 ){
		glColor3d(1.0, 0.0, 0.0);
	}
	else	glColor3d(0.2, 0.0, 0.0);
	glVertex3dv(robo_vertex[13]);

	if( robo_sencer[2] == 1 ){
		glColor3d(1.0, 0.0, 0.0);
	}
	else	glColor3d(0.2, 0.0, 0.0);
	glVertex3dv(robo_vertex[14]);

	glEnd();
	glPopMatrix();
}

void draw_ground(void){
	glColor3d(0.0, 0.5, 0.0);
	glBegin(GL_QUADS);
	glVertex3d(0.0, 0.0, 0.0);
	glVertex3d(170.0, 0.0, 0.0);
	glVertex3d(170.0, 170.0, 0.0);
	glVertex3d(0.0, 170.0, 0.0);
	glEnd();
}

void display(void)
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	//_̕ϊs̏
	glLoadIdentity();
	//_̌
	
	gluLookAt(
		85,85,500,
		85,85,0,
		0.0, 1.0, 0.0);

	rotation1 = 00;
	rotation2 = 0;

	glTranslated( 65.0, 65.0, 0.0);
	glRotated( rotation1, 0.0, 1.0, 0.0);
	glRotated( rotation2, 1.0, 0.0, 0.0);
	glTranslated( -65.0, -65.0, 0.0);

	draw_course();
	draw_robo();
	draw_ground();

	glutSwapBuffers();
}

void resize(int w, int h)
{
	GLdouble aspect;
	//ɂăr[|[gݒ肷
	glViewport(0, 0, w, h);
	//AXyNg̏
	aspect = (GLdouble)w/(GLdouble)h;
	//vWFNV[hŎˉe
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective( 20.0, aspect, 300.0, 700.0 );
	//m[}̃fr[[hֈڍs
	glMatrixMode( GL_MODELVIEW );
}


void keyboard(unsigned char key, int x, int y)
{
  if (key == '\033') exit(0);
}
///////////////////////////////////////////////////////////////////////////////
//sbNAbv
int pick(double x, double y){

	double wx, wy, wz;
	GLuint selectBuf[512];
	GLint hits;
	GLint viewport[4];
	float current_aspect;

	WindCo(x, y, 0.0, &wx, &wy, &wz);

	glGetIntegerv(GL_VIEWPORT, viewport);
	glSelectBuffer(512, selectBuf);
	(void) glRenderMode(GL_SELECT);

	glInitNames();
	glMatrixMode(GL_PROJECTION);			//vWFNV[h
	glPushMatrix();							//ZNV[ĥƂ̂݉̍svZ
		glLoadIdentity();					//𔽉f邽߂Push
		gluPickMatrix(wx,wy,		//sbLOs̏Z
			5.0,5.0, viewport);
		current_aspect = (float)viewport[2]/(float)viewport[3];
		gluPerspective( 20.0, current_aspect, 300.0, 700.0 );
		glMatrixMode( GL_MODELVIEW );		//Ńfr[[hɂIIII
		glPushMatrix();
		glLoadIdentity();
		gluLookAt(	85,85,500,	85,85,0, 0.0, 1.0, 0.0);
		glPushName(1);
		draw_course();
		glPopName();
		glPopMatrix();
		glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	glMatrixMode( GL_MODELVIEW );			//fr[[h֖߂

	hits = glRenderMode(GL_RENDER);
	if(hits != 0){
		return 1;
	}
	return 0;
}

///////////////////////////////////////////////////////////////////////////////
//}EX
void mouse(int button, int state, int x, int y){
	printf("%d %d \n",x,y);
	switch(button){
		case GLUT_LEFT_BUTTON:
			if(state == GLUT_DOWN);
			break;
	}
}

void idle(void)
{
	static int x_flag, y_flag;

	glutPostRedisplay();

	if(robo_x_position > 170.0 || robo_x_position < 0.0){
		if(x_flag == 0){
			robo_theta = PI - robo_theta;
			x_flag = 1;
		}
	}
	else	x_flag = 0;

	if(robo_y_position > 170.0 || robo_y_position < 0.0){
		if(y_flag == 0){
			robo_theta = - robo_theta;
			y_flag = 1;
		}
	}
	else	y_flag = 0;
	if( robo_sencer[1] == 1 && robo_sencer[0] == 0 && robo_sencer[2] == 0){
		robo_r_tire = 80;
		robo_l_tire = 80;
	}
	if( robo_sencer[0] == 1 && robo_sencer[1] == 0){
		robo_r_tire = 80;
		robo_l_tire = 10;
	}
	if( robo_sencer[0] == 1 && robo_sencer[1] == 1){
		robo_r_tire = 80;
		robo_l_tire = 30;
	}
	if( robo_sencer[2] == 1 && robo_sencer[1] == 0){
		robo_r_tire = 10;
		robo_l_tire = 80;
	}
	if( robo_sencer[2] == 1 && robo_sencer[1] == 1){
		robo_r_tire = 30;
		robo_l_tire = 80;
	}

	robo_omega = robo_tire_scale* ( robo_r_tire - robo_l_tire ) / 10.0;
	robo_theta += robo_omega;

	robo_velocity[0] = robo_tire_scale*( robo_r_tire + robo_l_tire )*cos(robo_theta);
	robo_velocity[1] = robo_tire_scale*( robo_r_tire + robo_l_tire )*sin(robo_theta);
	robo_x_position += robo_velocity[0];
	robo_y_position += robo_velocity[1];
}


///////////////////////////////////////////////////////////////////////////////
//OpenGLR}h
void init (void){
	glEnable(GL_DEPTH_TEST);
	//NCAg̈̓hԂF̐ݒ
	glClearColor( 0.0, 0.0, 0.0, 1.0 );
	//fvXobt@NAl̐ݒ
	glClearDepth( 1.0 );
}

int main(int argc, char *argv[])
{
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE);
  glutInitWindowSize(800, 640);
  glutInitWindowPosition(0, 0);
  glutCreateWindow("cɂ郍{bg@G~[^@Ver1.0");
  initialize_scene();
  initialize_robo();
  init();
  glClearColor(0.0, 0.0, 0.0, 0.0);
  glutDisplayFunc(display);
  glutKeyboardFunc(keyboard);
  glutReshapeFunc(resize);
  glutIdleFunc(idle); 
  glutMouseFunc(mouse);
  glutMainLoop();
}
/*----------------------------------------*/